/* -*- c++ -*- */
/*
 * Copyright 2004 Free Software Foundation, Inc.
 * 
 * This file is part of GNU Radio
 * 
 * GNU Radio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * GNU Radio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GNU Radio; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * config.h is generated by configure.  It contains the results
 * of probing for features, options etc.  It should be the first
 * file included in your .cc file.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gr_rds_data_decoder.h>
#include <gr_io_signature.h>

/*
 * Create a new instance of gr_rds_data_decoder and return
 * a boost shared_ptr.  This is effectively the public constructor.
 */
gr_rds_data_decoder_sptr 
gr_rds_make_data_decoder ()
	
{
  return gr_rds_data_decoder_sptr (new gr_rds_data_decoder ());
}

/*
 * Specify constraints on number of input and output streams.
 * This info is used to construct the input and output signatures
 * (2nd & 3rd args to gr_block's constructor).  The input and
 * output signatures are used by the runtime system to
 * check that a valid number and type of inputs and outputs
 * are connected to this block.  In this case, we accept
 * only 1 input and 1 output.
 */
static const int MIN_IN = 1;	// mininum number of input streams
static const int MAX_IN = 1;	// maximum number of input streams

/*
 * The private constructor
 */
gr_rds_data_decoder::gr_rds_data_decoder ()
  : gr_sync_block ("gr_rds_data_decoder",
	      gr_make_io_signature (MIN_IN, MAX_IN, sizeof (bool)),
	      gr_make_io_signature (0, 0, 0))
{

	reset_all_RDS_data();
	
	COLLECT_GROUPS_FOR_SYNC = 10;
	set_output_multiple(104); 	// Since a RDS datagroup contains 104 bits

	enter_no_sync();
}

/*
 * Our virtual destructor.
 */
gr_rds_data_decoder::~gr_rds_data_decoder ()
{}

bool gr_rds_data_decoder::check_data(const bool* block, offset_word_t offset)  {
	bool polynom[10] = {0,0,0,0,0, 0,0,0,0,0};
        bool bit = false;
	static bool offset_word[6][10] = 
	{{0, 0, 1, 1, 1, 1, 1, 1, 0, 0}, {0, 0, 0, 1, 1, 0, 0, 1, 1, 0},
	{0, 0, 0, 1, 0, 1, 1, 0, 1, 0}, {0, 0, 0, 0, 1, 0, 1, 0, 1, 1},
	{0, 0, 1, 0, 1, 1, 0, 1, 1, 0}, {0, 0, 0, 0, 0, 0, 0, 0, 0, 0}};

	// calc code polynom from data
	for(int i=0; i<16; i++ ) {
			bit = ( block[i] != polynom[9] );
                        for(int c=9; c > 0; c--) {
                                if((c==3) || (c==4) || (c==5) || (c==7) || (c==8)) {
                                        polynom[c] = (bit !=  polynom[c-1]);
                                } else {
                                        polynom[c] = polynom[c-1];                               
				}
                        }
                        polynom[0] = bit;
                }

                // add offset and check against received data
                for(int i=0; i < 10; i++) {
                        polynom[i] = (polynom[i] != offset_word[offset][i]);
			
                        if(block[25-i] != polynom[i]) 	{ 
				d_invalid_block_counter++;
				return false;
				
			}
                }
		return true;

	
}

void gr_rds_data_decoder::decode_PI(const bool* block) {
	// Remeber: input is indexed 0, 1, .... , 14, 15
	// but weight is 2^: 15, 14, 13, ... , 1, 0

	d_PI_CI = 0;
	d_PI_AC = 0;
	d_PI_PRN = 0;

	for(int i=0 ; i < 4; i++)
	{ if(block[i]) { d_PI_CI += (1<<(3-i)); } }

	for(int i = 4; i < 8; i++)
	{ if(block[i]) { d_PI_AC += (1<<(7-i)); } }

	for(int i= 8; i < 16; i++)
	{ if(block[i]) { d_PI_PRN += (1<<(15-i)); } }
	
	//printf("PI CI = %x, PI AC = %x PI_PRN = %x\n", d_PI_CI, d_PI_AC, d_PI_PRN);
	
}

int  gr_rds_data_decoder::decode_grouptype(const bool* block) {

	int group_type = 0;

	for(int i = 0; i < 4; i++) {
		if(block[i]) { group_type += (1<<3-i);}
	}

	if(!block[4]) { grouptypes[0][group_type] = true; }
	else { grouptypes[1][group_type] = true; }

	return group_type;
}

bool gr_rds_data_decoder::is_grouptype_A (const bool* block) {

	return !(block[4]);
}

void  gr_rds_data_decoder::decode_TP_PTY(const bool* block) {
	int pty = 0;

	for(int i = 6; i < 11; i++) {
		if(block[i]) { pty += (1<<4-i); }
	}

	d_PTY = pty;

	d_TP = block[5];
}

char gr_rds_data_decoder::transform_char(char z) {
	int row, col;
	static	char transform_table[16][8]  = {{' ', '0', '@', 'P', ' ', 'p', 'a', 'a'},
						{'!', '1', 'A', 'Q', 'a', 'q', 'a', 'a'},
						{'"', '2', 'B', 'R', 'b', 'r', 'e', 'e'},
						{'#', '3', 'C', 'S', 'c', 's', 'e', 'e'},
						{'', '4', 'D', 'T', 'd', 't', 'i', 'i'},
						{'%', '5', 'E', 'U', 'e', 'u', 'i', 'i'},
						{'&', '6', 'F', 'V', 'f', 'v', 'o', 'o'},
						{'\'', '7', 'G', 'W', 'g', 'w', 'o', 'o'},
						{'(', '8', 'H', 'X', 'h', 'x', 'u', 'u'},
						{')', '9', 'I', 'Y', 'i', 'y', 'u', 'u'},
						{'*', ':', 'J', 'Z', 'j', 'z', 'N', 'n'},
						{'+', ';', 'K', '[', 'k', '{', 'c', 'c'},
						{',', '<', 'L', '\\', 'l', '|', 's', 's'},
						{'-', '=', 'M', ']', 'm', '}', 'b', 'g'},
						{'.', '>', 'N', '-', 'n', '-', 'i', 'i'},
						{'/', '?', 'O', '-', 'o', ' ', 'I', 'i'}};

	row = (z & 0x0F);
	col = ((z >> 4) & 0x0F);
	col -= 2; // FIXME

	//printf("Char transform: row= %d, col= %d\n", row, col);

	return transform_table[row][col];
}

void  gr_rds_data_decoder::decode_PS_DI(const bool* block) {

	short seg_adr= 0;

	char ps0, ps1;
	ps0 = ps1 = 0;

	if(block[14])	{ seg_adr += 2;} 
	if(block[15])   { seg_adr += 1;}

	for(int i= 0; i < 8; i++) {
		if(block[i+52]) { ps0 += (1<<(7-i)); }
		if(block[i+60]) { ps1 += (1<<(7-i)) ;}
	}

	//printf("PS and DI seg_adr= %d, ps0= %d,  ps1=%d\n", seg_adr, ps0, ps1);
	ps0 = transform_char(ps0);
	ps1 = transform_char(ps1);

	switch(seg_adr) {
		case 0:	
			d_DI_Stereo = block[13];
			d_PS[0] = ps0;
			d_PS[1] = ps1;
			break;
		case 1:
			d_DI_AH = block[13];
			d_PS[2] = ps0;
			d_PS[3] = ps1;
			break;
		case 2:
			d_DI_Compressed = block[13];
			d_PS[4] = ps0;
			d_PS[5] = ps1;
			break;
		case 3:
			d_DI_dyn_PTY = block[13];
			d_PS[6] = ps0;
			d_PS[7] = ps1;
			break;
		default:
			return;
			
	}
}

void gr_rds_data_decoder::decode_AF(const bool* block) {
	// TODO
}

void gr_rds_data_decoder::decode_RT(const bool* block) {
	
	short txt_seg = 0;
	bool  a;
	char rt0, rt1, rt2, rt3;
	rt0 = rt1 = rt2 = rt3 = 0;

	a = block[11];

	for(int i= 0; i < 4; i++) {
		if(block[i+12]) { txt_seg += (1 << (3-i)); }
	}


	printf("RT: txt seg = %d, a/b = %d\n", txt_seg, a);


	for(int i=0; i < 8; i++) {
		if(block[i+26]) { rt0 += (1<<7-i); }
		if(block[i+34]) { rt1 += (1<<7-i); }
		if(block[i+52]) { rt2 += (1<<7-i); }
		if(block[i+60]) { rt3 += (1<<7-i); }
	}

	rt0 = transform_char(rt0);
	rt1 = transform_char(rt1);
	rt2 = transform_char(rt2);
	rt3 = transform_char(rt3);

	if(a){
		d_RT[0][txt_seg] = rt0;
		d_RT[0][txt_seg+1] = rt1;
		d_RT[0][txt_seg+2] = rt2;
		d_RT[0][txt_seg+3] = rt3;
	} else {
		 d_RT[1][txt_seg] = rt0;
                 d_RT[1][txt_seg+1] = rt1;
                 d_RT[1][txt_seg+2] = rt2;
                 d_RT[1][txt_seg+3] = rt3;
								 
	}

	printf("RT A: ");
	for(int i = 0; i < 64; i++) {
		printf("%c", d_RT[0][i]);
	}
	printf("\nRT B:");
	for(int i= 0; i < 64; i++) {
		printf("%c", d_RT[1][i]);
	}
	printf("\n");
	
	
}

void gr_rds_data_decoder::reset_all_RDS_data() {
	d_PI_CI = 0;
	d_PI_AC = 0;
	d_PI_PRN = 0;
	d_TP = false;
	d_TA = false;
	d_PTY = 0;
	d_DI_Stereo = d_DI_AH = d_DI_Compressed = d_DI_dyn_PTY = false;

	for(int i = 0; i < 16; i++) {
		grouptypes[0][i] = false;
		grouptypes[1][i] = false;
	}

	for(int i = 0; i < 8; i++) { d_PS[i] = ' '; }
//	d_AF.clear();
//	d_RT.clear();
	
	for(int i= 0; i < 64; i++) {
		d_RT[0][i] = d_RT[1][i] = ' ';
	}
}

void gr_rds_data_decoder::enter_no_sync() {
	
	for(int i = 0; i < 104; i++) {
		d_and[i] = true;
		d_nor[i]  = false;
		d_sync[i] = false;
	}

	d_collecting = 0;
	d_state = ST_NO_SYNC;
}

void gr_rds_data_decoder::enter_sync() {

	d_state = ST_SYNC;
}

int 
gr_rds_data_decoder::work (int noutput_items,
			       gr_vector_const_void_star &input_items,
			       gr_vector_void_star &output_items)
{
  const bool *in = (const bool *) input_items[0];

  //printf("RDS data decoder at work: input_items = %d, /104 = %d\n", noutput_items, noutput_items/104);

  int i;
  
/*  for ( i = 0; i < noutput_items; i++){
	if((i % 26) == 0) {printf(" "); }
	if((i % 26) % 16 == 0) { printf(" "); }
	if((i % 104) == 0) { printf ("\n");}
	printf("%d", in[i]);
	 
  }
  printf("\n");
  fflush(stdout);*/

  switch (d_state) {
	case ST_NO_SYNC:

		for(int offset=0; offset+103 < noutput_items; offset += 104) {
			for(i = 0; i < 104; i++) {
				d_and[i] = d_and[i] && in[i+offset];
				d_nor[i] = (d_nor[i]  || in[i+offset]);
				d_sync[i] = d_and[i] || !d_nor[i];

			}
			if(++d_collecting == COLLECT_GROUPS_FOR_SYNC) { 
				// ok enough groups 
				// check for a 26-wide-one-pattern
				//
				int oner=0;
				int first_occ=0;

/*				for(i = 0; i < 104; i++)
				{ printf("%d", d_sync[i]); }
				printf("\n");*/
				
			 	for(int k=0; k < 104; k++) {
					if(d_sync[k])
					{ 	
						if(oner == 0) { first_occ = k;}
						if(++oner == 26)
						{
						printf("\nSimple sync ok! offset k = %d\n", first_occ);
							enter_sync();
							return first_occ;
						}
					}
					else
					{ oner = 0;}
				}
				enter_no_sync();
				return noutput_items/2;
				
			}
		}
		return noutput_items;

	 	break; 
	case ST_SYNC:

		// Receiving a syncronized data stream
		//
	
/*		for ( i = 0; i < noutput_items; i++) {
			if((i % 26) == 0) {printf(" ");	}
			printf("%d", in[i]);
		}
		printf("\n");*/

		
		if(check_data(in, A)) { 
			// Data ok!
			// extract PI
			decode_PI(in);
//			printf("1. blk ok..\n");
		}

		if(check_data(in+26, B)) {
			// Data ok!
			// extract group type..
//			printf("2. blk ok..\n");

			int group_type = decode_grouptype(in+26);
			bool group_A = is_grouptype_A(in+26);

			decode_TP_PTY(in+26);

/*			printf("Group Type is: %d", group_type);
			if(group_A) {	printf("A\n"); }
			else { printf("B\n ");}*/

			switch (group_type) {
				case 0: 
					if(check_data(in+52, C) && check_data(in+78, D)) {
//						printf("3. & 4. blk ok!\n");
						
						decode_PS_DI(in+26);
						if(group_A) { decode_AF(in+52); }
		
						printf(" PS = ");
						for(int i = 0; i < 8; i++) {
							printf("%c", d_PS[i]);
						}
		
						if(d_DI_Stereo) { printf(" Stereo\n"); }
						else { printf(" Mono\n"); }
						break;
				case 2:
					if(check_data(in+52, C) && check_data(in+78, D)) {
//						printf("3. & 4. blk ok!\n");
						// Decode RT
						decode_RT(in+26);
					}
				}
			}
			
		}

		return noutput_items;					 


	default:
	  enter_no_sync();	// Error, invalid state!!!!
	  return 0;

  }
  // Tell runtime system how many output items we produced.
  return noutput_items;
}
