#!/usr/bin/env python
#
# Copyright 2004 Free Software Foundation, Inc.
# 
# This file is part of GNU Radio
# 
# GNU Radio is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
# 
# GNU Radio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with GNU Radio; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
# 

from gnuradio import gr
from gnuradio import mc4020
from gnuradio.eng_option import eng_option
from gnuradio.wxgui import stdgui, fftsink, scopesink
from optparse import OptionParser
import wx
import sys

def high_speed_adc (fg, input_rate):
    # return gr.file_source (gr.sizeof_short, "dummy.dat", False)
    return mc4020.source (input_rate, mc4020.MCC_CH3_EN | mc4020.MCC_ALL_1V)


def build_graph (fg, freq1, cfir_decimation):
    input_rate = 20e6
#    cfir_decimation = 100			# 200 hz
    audio_decimation = 1 

    quad_rate = input_rate / cfir_decimation
    audio_rate = quad_rate / audio_decimation

    
    # use high speed ADC as input source
    src = high_speed_adc (fg, input_rate)
    
    # compute FIR filter taps for channel selection
    channel_coeffs = \
      gr.firdes.low_pass (1,          # gain
                          input_rate,   # sampling rate
                          250e3,        # low pass cutoff freq
                          80e3,      # width of trans. band
                          gr.firdes.WIN_HAMMING)

    # input: short; output: complex
    chan_filter1 = \
      gr.freq_xlating_fir_filter_scf (cfir_decimation,
                                      channel_coeffs,
                                      freq1,        # 1st station freq
                                      input_rate)
    
    (head1, tail1) = build_pipeline (fg, quad_rate)
        
    # now wire it all together
    fg.connect (src, chan_filter1)
    fg.connect (chan_filter1, head1)
    
    return tail1 
    

def build_pipeline (fg, quad_rate):
    '''Given a flow_graph, fg, construct a pipeline
    for demodulating a broadcast FM signal.  The
    input is the downconverteed complex baseband
    signal. The output is the demodulated audio.

    build_pipeline returns a two element tuple
    containing the input and output endpoints.
    '''
    fm_demod_gain = 3200.0/32768.0
    volume = 1.0

    # input: complex; output: float
    fm_demod = gr.quadrature_demod_cf (volume*fm_demod_gain)

    # input: float; output: float
    audio_coeffs = gr.firdes.low_pass (20.0,            # gain
                                      quad_rate,      # sampling rate
                                      62e3,
                                      8e3,
                                      gr.firdes.WIN_HAMMING)
    audio_filter = gr.fir_filter_fff (1, audio_coeffs)

    fg.connect (fm_demod, audio_filter)
    return (fm_demod, audio_filter)
    


class app_flow_graph (stdgui.gui_flow_graph):
    def __init__(self, frame, panel, vbox, argv):
        stdgui.gui_flow_graph.__init__ (self, frame, panel, vbox, argv)

        
        parser = OptionParser (option_class=eng_option)
        (options, args) = parser.parse_args ()

	decimation = 100
	quad_rate =  20e6 / decimation # 200e3 
       
	
        scope_sink, scope_win = scopesink.make_scope_sink_f (self, panel, "SCOPE", quad_rate)
	fft_sink2, fft_win = fftsink.make_fft_sink_f(self, panel, "FFT", 1024, quad_rate, -60, 10)

	mixer = gr.multiply_ff()

	pilot_coeffs = gr.firdes.band_pass(5,
					  quad_rate,
					  18e3,
					  20e3,
					  2e3,
					  gr.firdes.WIN_HANN )
	pilot_filter = gr.fir_filter_fff(1, pilot_coeffs)

	rds_coeffs = gr.firdes.band_pass (10,
                                      quad_rate,
                                      54e3,
                                      60e3,
                                      3e3,
                                      gr.firdes.WIN_HANN)

        # input: float; output: float
	rds_filter = gr.fir_filter_fff (1, rds_coeffs)


	rds_bb_coeffs = gr.firdes.low_pass(1,
					   quad_rate,
					   1500,
					   3e3,
					   gr.firdes.WIN_HANN)

	rds_bb_filter = gr.fir_filter_fff(1, rds_bb_coeffs)

	tail = build_graph(self, 5.75e6, decimation)

	self.connect (tail, rds_filter)
	self.connect (tail, pilot_filter)

	mixer = gr.multiply_ff()

	self.connect(pilot_filter, (mixer, 0))
	self.connect(pilot_filter, (mixer, 1))
	self.connect(pilot_filter, (mixer, 2))
	self.connect(rds_filter, (mixer, 3))

	self.connect(mixer, rds_bb_filter)
	
	self.connect (rds_filter, (scope_sink,0))
	self.connect (rds_bb_filter, (scope_sink, 1))
        self.connect (tail,  fft_sink2)	
        vbox.Add (fft_win, 1, wx.EXPAND)
	vbox.Add (scope_win, 2 ,wx.EXPAND)


def main (args):

    nargs = len(args)
    if nargs == 1:
    	freq = float (args[0]) *1e6
    else:
    	print "Taking default freq: 88.85 (DRS3)"
	freq = 88.85e6

    print "Tuning freq"
    print freq
    print "\n"

    rf_front_end = gr.microtune_4937_eval_board ()
    if not rf_front_end.board_present_p ():
    	raise IOError, 'RF front end not found'

    rf_front_end.set_AGC(600)
    rf_front_end.set_RF_freq(freq)
    
    app = stdgui.stdapp (app_flow_graph, "RDS FFT")
    app.MainLoop ()

if __name__ == '__main__':
    main (sys.argv[1:])
